@@ -3,7 +3,7 @@
 /// @brief  The OmpLock and LockGuard classes are RAII-style
 ///         wrappers for OpenMP locks.
 ///
-/// Copyright (C) 2020 Kim Walisch, <kim.walisch@gmail.com>
+/// Copyright (C) 2022 Kim Walisch, <kim.walisch@gmail.com>
 ///
 /// This file is distributed under the BSD License. See the COPYING
 /// file in the top level directory.
@@ -12,6 +12,12 @@
 #ifndef OMPLOCK_HPP
 #define OMPLOCK_HPP
 
+#include <macros.hpp>
+
+#ifndef MAX_CACHE_LINE_SIZE
+  #define MAX_CACHE_LINE_SIZE 512
+#endif
+
 #if defined(_OPENMP)
   #include <omp.h>
 #else
@@ -35,25 +41,24 @@ namespace primecount {
 
 struct OmpLock
 {
-  OmpLock()
+  // The init() function should only be called
+  // when using >= 2 threads.
+  void init()
   {
+    enabled_ = true;
     omp_init_lock(&lock_);
   }
   ~OmpLock()
   {
-    omp_destroy_lock(&lock_);
-  }
-  char unused()
-  {
-    pad1[101] = 101;
-    pad2[102] = 102;
-    return pad1[101] | pad2[102];
+    if (enabled_)
+      omp_destroy_lock(&lock_);
   }
 
+  bool enabled_ = false;
   // Use padding to avoid CPU false sharing
-  char pad1[512];
+  MAYBE_UNUSED char pad1[MAX_CACHE_LINE_SIZE];
   omp_lock_t lock_;
-  char pad2[512];
+  MAYBE_UNUSED char pad2[MAX_CACHE_LINE_SIZE];
 };
 
 class LockGuard
@@ -61,16 +66,20 @@ class LockGuard
 public:
   LockGuard(OmpLock& lock)
   {
-    lock_ = &lock.lock_;
-    omp_set_lock(lock_);
+    if (lock.enabled_)
+    {
+      lock_ = &lock.lock_;
+      omp_set_lock(lock_);
+    }
   }
   ~LockGuard()
   {
-    omp_unset_lock(lock_);
+    if (lock_)
+      omp_unset_lock(lock_);
   }
 
 private:
-  omp_lock_t* lock_;
+  omp_lock_t* lock_ = nullptr;
 };
 
 } // namespace
