@@ -16,7 +16,7 @@
 ///        method, Revista do DETUA, vol. 4, no. 6, March 2006, p. 761.
 ///        http://sweet.ua.pt/tos/bib/5.4.pdf
 ///
-/// Copyright (C) 2021 Kim Walisch, <kim.walisch@gmail.com>
+/// Copyright (C) 2022 Kim Walisch, <kim.walisch@gmail.com>
 ///
 /// This file is distributed under the BSD License. See the COPYING
 /// file in the top level directory.
@@ -29,6 +29,7 @@
 #include <BitSieve240.hpp>
 #include <fast_div.hpp>
 #include <imath.hpp>
+#include <macros.hpp>
 #include <min.hpp>
 #include <PhiTiny.hpp>
 #include <PiTable.hpp>
@@ -36,13 +37,12 @@
 #include <popcnt.hpp>
 
 #include <stdint.h>
-#include <cassert>
+#include <algorithm>
 #include <utility>
 #include <vector>
 
 namespace {
 
-using namespace std;
 using namespace primecount;
 
 template <typename Primes>
@@ -92,7 +92,7 @@ public:
 
     // Make sure that there are no uninitialized
     // bits in the last sieve array element.
-    assert(max_x_size_ > 0);
+    ASSERT(max_x_size_ > 0);
     max_x_ = max_x_size_ * 240 - 1;
     max_a_ = max_a;
     sieve_.resize(max_a_ + 1);
@@ -128,7 +128,7 @@ public:
     else
     {
       c = c_cached;
-      assert(c_cached <= a);
+      ASSERT(c_cached <= a);
       sum = phi_cache(x, c) * SIGN;
     }
 
@@ -184,9 +184,6 @@ private:
 
   int64_t phi_cache(uint64_t x, uint64_t a) const
   {
-    assert(a < sieve_.size());
-    assert(x / 240 < sieve_[a].size());
-
     uint64_t count = sieve_[a][x / 240].count;
     uint64_t bits = sieve_[a][x / 240].bits;
     uint64_t bitmask = unset_larger_[x % 240];
@@ -219,14 +216,20 @@ private:
       // is not divisible by 2, 3 and 5. The 8 bits of each byte
       // correspond to the offsets { 1, 7, 11, 13, 17, 19, 23, 29 }.
       if (i == 3)
+      {
         sieve_[i].resize(max_x_size_);
+        std::fill(sieve_[i].begin(), sieve_[i].end(), sieve_t{0, ~0ull});
+      }
       else
       {
         // Initalize phi(x, i) with phi(x, i - 1)
         if (i - 1 <= PhiTiny::max_a())
           sieve_[i] = std::move(sieve_[i - 1]);
         else
-          sieve_[i] = sieve_[i - 1];
+        {
+          sieve_[i].resize(sieve_[i - 1].size());
+          std::copy(sieve_[i - 1].begin(), sieve_[i - 1].end(), sieve_[i].begin());
+        }
 
         // Remove prime[i] and its multiples
         uint64_t prime = primes_[i];
@@ -261,17 +264,16 @@ private:
   #pragma pack(push, 1)
   struct sieve_t
   {
-    uint32_t count = 0;
-    uint64_t bits = ~0ull;
+    uint32_t count;
+    uint64_t bits;
   };
-
   #pragma pack(pop)
 
   /// sieve[a] contains only numbers that are not divisible
   /// by any of the the first a primes. sieve[a][i].count
   /// contains the count of numbers < i * 240 that are not
   /// divisible by any of the first a primes.
-  vector<vector<sieve_t>> sieve_;
+  std::vector<pod_vector<sieve_t>> sieve_;
   const Primes& primes_;
   const PiTable& pi_;
 };
