@@ -22,13 +22,13 @@
 ///        speedup counting. However using a binary indexed tree is
 ///        bad for performance as it causes many cache misses and
 ///        branch mispredictions. For this reason this implementation
-///        instead uses a counters array whose elements contain the
-///        total count of unsieved elements in a certain interval.
+///        instead uses a linear counter array whose elements contain
+///        the total count of unsieved elements in a certain interval.
 ///
 ///        In-depth description of this algorithm:
 ///        https://github.com/kimwalisch/primecount/blob/master/doc/Hard-Special-Leaves.md
 ///
-/// Copyright (C) 2021 Kim Walisch, <kim.walisch@gmail.com>
+/// Copyright (C) 2022 Kim Walisch, <kim.walisch@gmail.com>
 ///
 /// This file is distributed under the BSD License. See the COPYING
 /// file in the top level directory.
@@ -39,16 +39,15 @@
 #include <imath.hpp>
 #include <macros.hpp>
 #include <min.hpp>
+#include <pod_vector.hpp>
 #include <popcnt.hpp>
 
 #include <stdint.h>
 #include <algorithm>
-#include <array>
-#include <cassert>
-#include <vector>
 
-using namespace std;
-using namespace primecount;
+using std::fill_n;
+using std::sqrt;
+using primecount::pod_array;
 
 namespace {
 
@@ -61,7 +60,7 @@ struct WheelInit
 /// Categorize sieving primes according to their modulo 30
 /// congruence class { 1, 7, 11, 13, 17, 19, 23, 29 }.
 ///
-const array<uint8_t, 30> wheel_offsets =
+const pod_array<uint8_t, 30> wheel_offsets =
 {
   0, 8 * 0, 0, 0, 0, 0,
   0, 8 * 1, 0, 0, 0, 8 * 2,
@@ -73,7 +72,7 @@ const array<uint8_t, 30> wheel_offsets =
 /// Used to calculate the first multiple > start of a
 /// sieving prime that is coprime to 2, 3, 5.
 ///
-const array<WheelInit, 30> wheel_init
+const pod_array<WheelInit, 30> wheel_init
 {{
   {1,  0}, {0,  0}, {5,  1}, {4,  1}, {3,  1},
   {2,  1}, {1,  1}, {0,  1}, {3,  2}, {2,  2},
@@ -91,8 +90,8 @@ Sieve::Sieve(uint64_t low,
              uint64_t segment_size, 
              uint64_t wheel_size)
 {
-  assert(low % 30 == 0);
-  assert(segment_size % 240 == 0);
+  ASSERT(low % 30 == 0);
+  ASSERT(segment_size % 240 == 0);
 
   start_ = low;
   segment_size = get_segment_size(segment_size);
@@ -101,16 +100,15 @@ Sieve::Sieve(uint64_t low,
   // to 30 numbers i.e. the 8 bits correspond to the
   // offsets = {1, 7, 11, 13, 17, 19, 23, 29}.
   sieve_.resize(segment_size / 30);
-
   wheel_.reserve(wheel_size);
   wheel_.resize(4);
-  allocate_counters(low);
+  allocate_counter(low);
 }
 
-/// Each element of the counters array contains the current
+/// Each element of the counter array contains the current
 /// number of unsieved elements in the interval:
-/// [i * counters_dist, (i + 1) * counters_dist[.
-/// Ideally each element of the counters array should
+/// [i * counter_.dist, (i + 1) * counter_.dist[.
+/// Ideally each element of the counter array should
 /// represent an interval of size O(sqrt(average_leaf_dist)).
 /// Also the counter distance should be adjusted regularly
 /// whilst sieving as the distance between consecutive
@@ -118,34 +116,40 @@ Sieve::Sieve(uint64_t low,
 /// sieving algorithm but grows up to segment_size towards
 /// the end of the algorithm.
 ///
-void Sieve::allocate_counters(uint64_t low)
+void Sieve::allocate_counter(uint64_t low)
 {
   double average_leaf_dist = sqrt(low);
-  double counters_dist = sqrt(average_leaf_dist);
+  double counter_dist = sqrt(average_leaf_dist);
 
-  // Here we balance counting with the counters array and
+  // Here we balance counting with the counter array and
   // counting from the sieve array using the POPCNT
   // instruction. Since the POPCNT instructions allows to
   // count a distance of 240 using a single instruction we
   // slightly increase the counter distance and slightly
-  // decrease the size of the counters array.
-  counters_dist_ = (uint64_t) (counters_dist * sqrt(240));
+  // decrease the size of the counter array.
+  counter_.dist = (uint64_t) (counter_dist * sqrt(240));
 
   // Each byte represents an interval of size 30
-  uint64_t byte_dist = counters_dist_ / 30;
-  byte_dist = max(byte_dist, 64);
-  byte_dist = next_power_of_2(byte_dist);
+  uint64_t bytes = counter_.dist / 30;
 
-  // Make sure the counters (32-bit) don't overflow.
-  // This can never happen since each counters array element
+  // Increasing the minimum counter distance decreases the
+  // branch mispredictions but on the other hand increases
+  // the number of executed instructions. On newer CPUs
+  // reducing the branch mispredictions is more important
+  // than reducing the number of executed instructions.
+  bytes = max((uint64_t) sizeof(uint64_t) * 16, bytes);
+  bytes = next_power_of_2(bytes);
+
+  // Make sure the counter (32-bit) doesn't overflow.
+  // This can never happen since each counter array element
   // only counts the number of unsieved elements (1 bits) in
   // an interval of size: sieve_limit^(1/4) * sqrt(240).
   // Hence the max(counter value) = 2^18.
-  assert(byte_dist * 8 <= numeric_limits<uint32_t>::max());
-  uint64_t counters_size = ceil_div(sieve_.size(), byte_dist);
-  counters_.resize(counters_size);
-  counters_dist_ = byte_dist * 30;
-  counters_dist_log2_ = ilog2(byte_dist);
+  ASSERT(bytes * 8 <= std::numeric_limits<uint32_t>::max());
+  uint64_t counter_size = ceil_div(sieve_.size(), bytes);
+  counter_.counter.resize(counter_size);
+  counter_.dist = bytes * 30;
+  counter_.log2_dist = ilog2(bytes);
 }
 
 /// The segment size is sieve.size() * 30 as each
@@ -162,7 +166,7 @@ uint64_t Sieve::segment_size() const
 ///
 uint64_t Sieve::get_segment_size(uint64_t size)
 {
-  size = max<uint64_t>(240, size);
+  size = max(size, 240);
 
   if (size % 240)
     size += 240 - size % 240;
@@ -185,18 +189,18 @@ void Sieve::reset_sieve(uint64_t low, uint64_t high)
   }
 }
 
-void Sieve::reset_counters()
+void Sieve::reset_counter()
 {
   prev_stop_ = 0;
   count_ = 0;
-  counters_i_ = 0;
-  counters_count_ = 0;
-  counters_stop_ = counters_dist_;
+  counter_.i = 0;
+  counter_.sum = 0;
+  counter_.stop = counter_.dist;
 }
 
-void Sieve::init_counters(uint64_t low, uint64_t high)
+void Sieve::init_counter(uint64_t low, uint64_t high)
 {
-  reset_counters();
+  reset_counter();
   total_count_ = 0;
 
   uint64_t start = 0;
@@ -204,42 +208,42 @@ void Sieve::init_counters(uint64_t low, uint64_t high)
 
   while (start <= max_stop)
   {
-    uint64_t stop = start + counters_dist_ - 1;
+    uint64_t stop = start + counter_.dist - 1;
     stop = min(stop, max_stop);
     uint64_t cnt = count(start, stop);
     uint64_t byte_index = start / 30;
-    uint64_t i = byte_index >> counters_dist_log2_;
+    uint64_t i = byte_index >> counter_.log2_dist;
 
-    counters_[i] = (uint32_t) cnt;
+    counter_[i] = (uint32_t) cnt;
     total_count_ += cnt;
-    start += counters_dist_;
+    start += counter_.dist;
   }
 }
 
 /// Count 1 bits inside [0, stop]
 uint64_t Sieve::count(uint64_t stop)
 {
-  assert(stop >= prev_stop_);
+  ASSERT(stop >= prev_stop_);
   uint64_t start = prev_stop_ + 1;
   prev_stop_ = stop;
 
   // Quickly count the number of unsieved elements (in
   // the sieve array) up to a value that is close to
-  // the stop number i.e. (stop - value) <= counters_dist.
-  // We do this using the counters array, each element
-  // of the counters array contains the number of
+  // the stop number i.e. (stop - start) < counter_.dist.
+  // We do this using the counter array, each element
+  // of the counter array contains the number of
   // unsieved elements in the interval:
-  // [i * counters_dist, (i + 1) * counters_dist[.
-  while (counters_stop_ <= stop)
+  // [i * counter_.dist, (i + 1) * counter_.dist[.
+  while (counter_.stop <= stop)
   {
-    start = counters_stop_;
-    counters_stop_ += counters_dist_;
-    counters_count_ += counters_[counters_i_++];
-    count_ = counters_count_;
+    start = counter_.stop;
+    counter_.stop += counter_.dist;
+    counter_.sum += counter_[counter_.i++];
+    count_ = counter_.sum;
   }
 
   // Here the remaining distance is relatively small i.e.
-  // (stop - start) <= counters_dist, hence we simply
+  // (stop - start) < counter_.dist, hence we simply
   // count the remaining number of unsieved elements by
   // linearly iterating over the sieve array.
   count_ += count(start, stop);
@@ -252,7 +256,7 @@ uint64_t Sieve::count(uint64_t start, uint64_t stop) const
   if (start > stop)
     return 0;
 
-  assert(stop - start < segment_size());
+  ASSERT(stop - start < segment_size());
 
   uint64_t start_idx = start / 240;
   uint64_t stop_idx = stop / 240;
@@ -278,7 +282,7 @@ uint64_t Sieve::count(uint64_t start, uint64_t stop) const
 ///
 void Sieve::add(uint64_t prime)
 {
-  assert(start_ % 30 == 0);
+  ASSERT(start_ % 30 == 0);
 
   // first multiple > start_
   uint64_t quotient = start_ / prime + 1;
@@ -323,7 +327,23 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
   {
     for (;;)
     {
-      case 0: CHECK_FINISHED(0); sieve[m] &= ~(1 << 0); m += prime * 6 + 0; FALLTHROUGH;
+      case 0: {
+                uint64_t max_offset = m + prime * 28;
+                uint64_t limit = std::max(max_offset, sieve_size) - max_offset;
+
+                for (; m < limit; m += prime * 30 + 1)
+                {
+                  sieve[m + prime *  0] &= ~(1 << 0);
+                  sieve[m + prime *  6] &= ~(1 << 1);
+                  sieve[m + prime * 10] &= ~(1 << 2);
+                  sieve[m + prime * 12] &= ~(1 << 3);
+                  sieve[m + prime * 16] &= ~(1 << 4);
+                  sieve[m + prime * 18] &= ~(1 << 5);
+                  sieve[m + prime * 22] &= ~(1 << 6);
+                  sieve[m + prime * 28] &= ~(1 << 7);
+                }
+              }
+              CHECK_FINISHED(0); sieve[m] &= ~(1 << 0); m += prime * 6 + 0; FALLTHROUGH;
       case 1: CHECK_FINISHED(1); sieve[m] &= ~(1 << 1); m += prime * 4 + 0; FALLTHROUGH;
       case 2: CHECK_FINISHED(2); sieve[m] &= ~(1 << 2); m += prime * 2 + 0; FALLTHROUGH;
       case 3: CHECK_FINISHED(3); sieve[m] &= ~(1 << 3); m += prime * 4 + 0; FALLTHROUGH;
@@ -331,24 +351,27 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
       case 5: CHECK_FINISHED(5); sieve[m] &= ~(1 << 5); m += prime * 4 + 0; FALLTHROUGH;
       case 6: CHECK_FINISHED(6); sieve[m] &= ~(1 << 6); m += prime * 6 + 0; FALLTHROUGH;
       case 7: CHECK_FINISHED(7); sieve[m] &= ~(1 << 7); m += prime * 2 + 1;
-
-      while (m + prime * 28 < sieve_size)
-      {
-        sieve[m + prime *  0] &= ~(1 << 0);
-        sieve[m + prime *  6] &= ~(1 << 1);
-        sieve[m + prime * 10] &= ~(1 << 2);
-        sieve[m + prime * 12] &= ~(1 << 3);
-        sieve[m + prime * 16] &= ~(1 << 4);
-        sieve[m + prime * 18] &= ~(1 << 5);
-        sieve[m + prime * 22] &= ~(1 << 6);
-        sieve[m + prime * 28] &= ~(1 << 7);
-        m += prime * 30 + 1;
-      }
     }
 
     for (;;)
     {
-      case  8: CHECK_FINISHED( 8); sieve[m] &= ~(1 << 1); m += prime * 6 + 1; FALLTHROUGH;
+      case  8: {
+                 uint64_t max_offset = m + prime * 28 + 6;
+                 uint64_t limit = std::max(max_offset, sieve_size) - max_offset;
+
+                 for (; m < limit; m += prime * 30 + 7)
+                 {
+                   sieve[m + prime *  0 + 0] &= ~(1 << 1);
+                   sieve[m + prime *  6 + 1] &= ~(1 << 5);
+                   sieve[m + prime * 10 + 2] &= ~(1 << 4);
+                   sieve[m + prime * 12 + 3] &= ~(1 << 0);
+                   sieve[m + prime * 16 + 3] &= ~(1 << 7);
+                   sieve[m + prime * 18 + 4] &= ~(1 << 3);
+                   sieve[m + prime * 22 + 5] &= ~(1 << 2);
+                   sieve[m + prime * 28 + 6] &= ~(1 << 6);
+                 }
+               }
+               CHECK_FINISHED( 8); sieve[m] &= ~(1 << 1); m += prime * 6 + 1; FALLTHROUGH;
       case  9: CHECK_FINISHED( 9); sieve[m] &= ~(1 << 5); m += prime * 4 + 1; FALLTHROUGH;
       case 10: CHECK_FINISHED(10); sieve[m] &= ~(1 << 4); m += prime * 2 + 1; FALLTHROUGH;
       case 11: CHECK_FINISHED(11); sieve[m] &= ~(1 << 0); m += prime * 4 + 0; FALLTHROUGH;
@@ -356,24 +379,27 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
       case 13: CHECK_FINISHED(13); sieve[m] &= ~(1 << 3); m += prime * 4 + 1; FALLTHROUGH;
       case 14: CHECK_FINISHED(14); sieve[m] &= ~(1 << 2); m += prime * 6 + 1; FALLTHROUGH;
       case 15: CHECK_FINISHED(15); sieve[m] &= ~(1 << 6); m += prime * 2 + 1;
-
-      while (m + prime * 28 + 6 < sieve_size)
-      {
-        sieve[m + prime *  0 + 0] &= ~(1 << 1);
-        sieve[m + prime *  6 + 1] &= ~(1 << 5);
-        sieve[m + prime * 10 + 2] &= ~(1 << 4);
-        sieve[m + prime * 12 + 3] &= ~(1 << 0);
-        sieve[m + prime * 16 + 3] &= ~(1 << 7);
-        sieve[m + prime * 18 + 4] &= ~(1 << 3);
-        sieve[m + prime * 22 + 5] &= ~(1 << 2);
-        sieve[m + prime * 28 + 6] &= ~(1 << 6);
-        m += prime * 30 + 7;
-      }
     }
 
     for (;;)
     {
-      case 16: CHECK_FINISHED(16); sieve[m] &= ~(1 << 2); m += prime * 6 + 2; FALLTHROUGH;
+      case 16: {
+                 uint64_t max_offset = m + prime * 28 + 10;
+                 uint64_t limit = std::max(max_offset, sieve_size) - max_offset;
+
+                 for (; m < limit; m += prime * 30 + 11)
+                 {
+                   sieve[m + prime *  0 +  0] &= ~(1 << 2);
+                   sieve[m + prime *  6 +  2] &= ~(1 << 4);
+                   sieve[m + prime * 10 +  4] &= ~(1 << 0);
+                   sieve[m + prime * 12 +  4] &= ~(1 << 6);
+                   sieve[m + prime * 16 +  6] &= ~(1 << 1);
+                   sieve[m + prime * 18 +  6] &= ~(1 << 7);
+                   sieve[m + prime * 22 +  8] &= ~(1 << 3);
+                   sieve[m + prime * 28 + 10] &= ~(1 << 5);
+                 }
+               }
+               CHECK_FINISHED(16); sieve[m] &= ~(1 << 2); m += prime * 6 + 2; FALLTHROUGH;
       case 17: CHECK_FINISHED(17); sieve[m] &= ~(1 << 4); m += prime * 4 + 2; FALLTHROUGH;
       case 18: CHECK_FINISHED(18); sieve[m] &= ~(1 << 0); m += prime * 2 + 0; FALLTHROUGH;
       case 19: CHECK_FINISHED(19); sieve[m] &= ~(1 << 6); m += prime * 4 + 2; FALLTHROUGH;
@@ -381,24 +407,27 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
       case 21: CHECK_FINISHED(21); sieve[m] &= ~(1 << 7); m += prime * 4 + 2; FALLTHROUGH;
       case 22: CHECK_FINISHED(22); sieve[m] &= ~(1 << 3); m += prime * 6 + 2; FALLTHROUGH;
       case 23: CHECK_FINISHED(23); sieve[m] &= ~(1 << 5); m += prime * 2 + 1;
-
-      while (m + prime * 28 + 10 < sieve_size)
-      {
-        sieve[m + prime *  0 +  0] &= ~(1 << 2);
-        sieve[m + prime *  6 +  2] &= ~(1 << 4);
-        sieve[m + prime * 10 +  4] &= ~(1 << 0);
-        sieve[m + prime * 12 +  4] &= ~(1 << 6);
-        sieve[m + prime * 16 +  6] &= ~(1 << 1);
-        sieve[m + prime * 18 +  6] &= ~(1 << 7);
-        sieve[m + prime * 22 +  8] &= ~(1 << 3);
-        sieve[m + prime * 28 + 10] &= ~(1 << 5);
-        m += prime * 30 + 11;
-      }
     }
 
     for (;;)
     {
-      case 24: CHECK_FINISHED(24); sieve[m] &= ~(1 << 3); m += prime * 6 + 3; FALLTHROUGH;
+      case 24: {
+                 uint64_t max_offset = m + prime * 28 + 12;
+                 uint64_t limit = std::max(max_offset, sieve_size) - max_offset;
+
+                 for (; m < limit; m += prime * 30 + 13)
+                 {
+                   sieve[m + prime *  0 +  0] &= ~(1 << 3);
+                   sieve[m + prime *  6 +  3] &= ~(1 << 0);
+                   sieve[m + prime * 10 +  4] &= ~(1 << 6);
+                   sieve[m + prime * 12 +  5] &= ~(1 << 5);
+                   sieve[m + prime * 16 +  7] &= ~(1 << 2);
+                   sieve[m + prime * 18 +  8] &= ~(1 << 1);
+                   sieve[m + prime * 22 +  9] &= ~(1 << 7);
+                   sieve[m + prime * 28 + 12] &= ~(1 << 4);
+                 }
+               }
+               CHECK_FINISHED(24); sieve[m] &= ~(1 << 3); m += prime * 6 + 3; FALLTHROUGH;
       case 25: CHECK_FINISHED(25); sieve[m] &= ~(1 << 0); m += prime * 4 + 1; FALLTHROUGH;
       case 26: CHECK_FINISHED(26); sieve[m] &= ~(1 << 6); m += prime * 2 + 1; FALLTHROUGH;
       case 27: CHECK_FINISHED(27); sieve[m] &= ~(1 << 5); m += prime * 4 + 2; FALLTHROUGH;
@@ -406,24 +435,27 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
       case 29: CHECK_FINISHED(29); sieve[m] &= ~(1 << 1); m += prime * 4 + 1; FALLTHROUGH;
       case 30: CHECK_FINISHED(30); sieve[m] &= ~(1 << 7); m += prime * 6 + 3; FALLTHROUGH;
       case 31: CHECK_FINISHED(31); sieve[m] &= ~(1 << 4); m += prime * 2 + 1;
-
-      while (m + prime * 28 + 12 < sieve_size)
-      {
-        sieve[m + prime *  0 +  0] &= ~(1 << 3);
-        sieve[m + prime *  6 +  3] &= ~(1 << 0);
-        sieve[m + prime * 10 +  4] &= ~(1 << 6);
-        sieve[m + prime * 12 +  5] &= ~(1 << 5);
-        sieve[m + prime * 16 +  7] &= ~(1 << 2);
-        sieve[m + prime * 18 +  8] &= ~(1 << 1);
-        sieve[m + prime * 22 +  9] &= ~(1 << 7);
-        sieve[m + prime * 28 + 12] &= ~(1 << 4);
-        m += prime * 30 + 13;
-      }
     }
 
     for (;;)
     {
-      case 32: CHECK_FINISHED(32); sieve[m] &= ~(1 << 4); m += prime * 6 + 3; FALLTHROUGH;
+      case 32: {
+                 uint64_t max_offset = m + prime * 28 + 16;
+                 uint64_t limit = std::max(max_offset, sieve_size) - max_offset;
+
+                 for (; m < limit; m += prime * 30 + 17)
+                 {
+                   sieve[m + prime *  0 +  0] &= ~(1 << 4);
+                   sieve[m + prime *  6 +  3] &= ~(1 << 7);
+                   sieve[m + prime * 10 +  6] &= ~(1 << 1);
+                   sieve[m + prime * 12 +  7] &= ~(1 << 2);
+                   sieve[m + prime * 16 +  9] &= ~(1 << 5);
+                   sieve[m + prime * 18 + 10] &= ~(1 << 6);
+                   sieve[m + prime * 22 + 13] &= ~(1 << 0);
+                   sieve[m + prime * 28 + 16] &= ~(1 << 3);
+                 }
+               }
+               CHECK_FINISHED(32); sieve[m] &= ~(1 << 4); m += prime * 6 + 3; FALLTHROUGH;
       case 33: CHECK_FINISHED(33); sieve[m] &= ~(1 << 7); m += prime * 4 + 3; FALLTHROUGH;
       case 34: CHECK_FINISHED(34); sieve[m] &= ~(1 << 1); m += prime * 2 + 1; FALLTHROUGH;
       case 35: CHECK_FINISHED(35); sieve[m] &= ~(1 << 2); m += prime * 4 + 2; FALLTHROUGH;
@@ -431,24 +463,27 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
       case 37: CHECK_FINISHED(37); sieve[m] &= ~(1 << 6); m += prime * 4 + 3; FALLTHROUGH;
       case 38: CHECK_FINISHED(38); sieve[m] &= ~(1 << 0); m += prime * 6 + 3; FALLTHROUGH;
       case 39: CHECK_FINISHED(39); sieve[m] &= ~(1 << 3); m += prime * 2 + 1;
-
-      while (m + prime * 28 + 16 < sieve_size)
-      {
-        sieve[m + prime *  0 +  0] &= ~(1 << 4);
-        sieve[m + prime *  6 +  3] &= ~(1 << 7);
-        sieve[m + prime * 10 +  6] &= ~(1 << 1);
-        sieve[m + prime * 12 +  7] &= ~(1 << 2);
-        sieve[m + prime * 16 +  9] &= ~(1 << 5);
-        sieve[m + prime * 18 + 10] &= ~(1 << 6);
-        sieve[m + prime * 22 + 13] &= ~(1 << 0);
-        sieve[m + prime * 28 + 16] &= ~(1 << 3);
-        m += prime * 30 + 17;
-      }
     }
 
     for (;;)
     {
-      case 40: CHECK_FINISHED(40); sieve[m] &= ~(1 << 5); m += prime * 6 + 4; FALLTHROUGH;
+      case 40: {
+                 uint64_t max_offset = m + prime * 28 + 18;
+                 uint64_t limit = std::max(max_offset, sieve_size) - max_offset;
+
+                 for (; m < limit; m += prime * 30 + 19)
+                 {
+                   sieve[m + prime *  0 +  0] &= ~(1 << 5);
+                   sieve[m + prime *  6 +  4] &= ~(1 << 3);
+                   sieve[m + prime * 10 +  6] &= ~(1 << 7);
+                   sieve[m + prime * 12 +  8] &= ~(1 << 1);
+                   sieve[m + prime * 16 + 10] &= ~(1 << 6);
+                   sieve[m + prime * 18 + 12] &= ~(1 << 0);
+                   sieve[m + prime * 22 + 14] &= ~(1 << 4);
+                   sieve[m + prime * 28 + 18] &= ~(1 << 2);
+                 }
+               }
+               CHECK_FINISHED(40); sieve[m] &= ~(1 << 5); m += prime * 6 + 4; FALLTHROUGH;
       case 41: CHECK_FINISHED(41); sieve[m] &= ~(1 << 3); m += prime * 4 + 2; FALLTHROUGH;
       case 42: CHECK_FINISHED(42); sieve[m] &= ~(1 << 7); m += prime * 2 + 2; FALLTHROUGH;
       case 43: CHECK_FINISHED(43); sieve[m] &= ~(1 << 1); m += prime * 4 + 2; FALLTHROUGH;
@@ -456,24 +491,27 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
       case 45: CHECK_FINISHED(45); sieve[m] &= ~(1 << 0); m += prime * 4 + 2; FALLTHROUGH;
       case 46: CHECK_FINISHED(46); sieve[m] &= ~(1 << 4); m += prime * 6 + 4; FALLTHROUGH;
       case 47: CHECK_FINISHED(47); sieve[m] &= ~(1 << 2); m += prime * 2 + 1;
-
-      while (m + prime * 28 + 18 < sieve_size)
-      {
-        sieve[m + prime *  0 +  0] &= ~(1 << 5);
-        sieve[m + prime *  6 +  4] &= ~(1 << 3);
-        sieve[m + prime * 10 +  6] &= ~(1 << 7);
-        sieve[m + prime * 12 +  8] &= ~(1 << 1);
-        sieve[m + prime * 16 + 10] &= ~(1 << 6);
-        sieve[m + prime * 18 + 12] &= ~(1 << 0);
-        sieve[m + prime * 22 + 14] &= ~(1 << 4);
-        sieve[m + prime * 28 + 18] &= ~(1 << 2);
-        m += prime * 30 + 19;
-      }
     }
 
     for (;;)
     {
-      case 48: CHECK_FINISHED(48); sieve[m] &= ~(1 << 6); m += prime * 6 + 5; FALLTHROUGH;
+      case 48: {
+                 uint64_t max_offset = m + prime * 28 + 22;
+                 uint64_t limit = std::max(max_offset, sieve_size) - max_offset;
+
+                 for (; m < limit; m += prime * 30 + 23)
+                 {
+                   sieve[m + prime *  0 +  0] &= ~(1 << 6);
+                   sieve[m + prime *  6 +  5] &= ~(1 << 2);
+                   sieve[m + prime * 10 +  8] &= ~(1 << 3);
+                   sieve[m + prime * 12 +  9] &= ~(1 << 7);
+                   sieve[m + prime * 16 + 13] &= ~(1 << 0);
+                   sieve[m + prime * 18 + 14] &= ~(1 << 4);
+                   sieve[m + prime * 22 + 17] &= ~(1 << 5);
+                   sieve[m + prime * 28 + 22] &= ~(1 << 1);
+                 }
+               }
+               CHECK_FINISHED(48); sieve[m] &= ~(1 << 6); m += prime * 6 + 5; FALLTHROUGH;
       case 49: CHECK_FINISHED(49); sieve[m] &= ~(1 << 2); m += prime * 4 + 3; FALLTHROUGH;
       case 50: CHECK_FINISHED(50); sieve[m] &= ~(1 << 3); m += prime * 2 + 1; FALLTHROUGH;
       case 51: CHECK_FINISHED(51); sieve[m] &= ~(1 << 7); m += prime * 4 + 4; FALLTHROUGH;
@@ -481,24 +519,27 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
       case 53: CHECK_FINISHED(53); sieve[m] &= ~(1 << 4); m += prime * 4 + 3; FALLTHROUGH;
       case 54: CHECK_FINISHED(54); sieve[m] &= ~(1 << 5); m += prime * 6 + 5; FALLTHROUGH;
       case 55: CHECK_FINISHED(55); sieve[m] &= ~(1 << 1); m += prime * 2 + 1;
-
-      while (m + prime * 28 + 22 < sieve_size)
-      {
-        sieve[m + prime *  0 +  0] &= ~(1 << 6);
-        sieve[m + prime *  6 +  5] &= ~(1 << 2);
-        sieve[m + prime * 10 +  8] &= ~(1 << 3);
-        sieve[m + prime * 12 +  9] &= ~(1 << 7);
-        sieve[m + prime * 16 + 13] &= ~(1 << 0);
-        sieve[m + prime * 18 + 14] &= ~(1 << 4);
-        sieve[m + prime * 22 + 17] &= ~(1 << 5);
-        sieve[m + prime * 28 + 22] &= ~(1 << 1);
-        m += prime * 30 + 23;
-      }
     }
 
     for (;;)
     {
-      case 56: CHECK_FINISHED(56); sieve[m] &= ~(1 << 7); m += prime * 6 + 6; FALLTHROUGH;
+      case 56: {
+                 uint64_t max_offset = m + prime * 28 + 28;
+                 uint64_t limit = std::max(max_offset, sieve_size) - max_offset;
+
+                 for (; m < limit; m += prime * 30 + 29)
+                 {
+                   sieve[m + prime *  0 +  0] &= ~(1 << 7);
+                   sieve[m + prime *  6 +  6] &= ~(1 << 6);
+                   sieve[m + prime * 10 + 10] &= ~(1 << 5);
+                   sieve[m + prime * 12 + 12] &= ~(1 << 4);
+                   sieve[m + prime * 16 + 16] &= ~(1 << 3);
+                   sieve[m + prime * 18 + 18] &= ~(1 << 2);
+                   sieve[m + prime * 22 + 22] &= ~(1 << 1);
+                   sieve[m + prime * 28 + 28] &= ~(1 << 0);
+                 }
+               }
+               CHECK_FINISHED(56); sieve[m] &= ~(1 << 7); m += prime * 6 + 6; FALLTHROUGH;
       case 57: CHECK_FINISHED(57); sieve[m] &= ~(1 << 6); m += prime * 4 + 4; FALLTHROUGH;
       case 58: CHECK_FINISHED(58); sieve[m] &= ~(1 << 5); m += prime * 2 + 2; FALLTHROUGH;
       case 59: CHECK_FINISHED(59); sieve[m] &= ~(1 << 4); m += prime * 4 + 4; FALLTHROUGH;
@@ -506,19 +547,6 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
       case 61: CHECK_FINISHED(61); sieve[m] &= ~(1 << 2); m += prime * 4 + 4; FALLTHROUGH;
       case 62: CHECK_FINISHED(62); sieve[m] &= ~(1 << 1); m += prime * 6 + 6; FALLTHROUGH;
       case 63: CHECK_FINISHED(63); sieve[m] &= ~(1 << 0); m += prime * 2 + 1;
-
-      while (m + prime * 28 + 28 < sieve_size)
-      {
-        sieve[m + prime *  0 +  0] &= ~(1 << 7);
-        sieve[m + prime *  6 +  6] &= ~(1 << 6);
-        sieve[m + prime * 10 + 10] &= ~(1 << 5);
-        sieve[m + prime * 12 + 12] &= ~(1 << 4);
-        sieve[m + prime * 16 + 16] &= ~(1 << 3);
-        sieve[m + prime * 18 + 18] &= ~(1 << 2);
-        sieve[m + prime * 22 + 22] &= ~(1 << 1);
-        sieve[m + prime * 28 + 28] &= ~(1 << 0);
-        m += prime * 30 + 29;
-      }
     }
 
     default: UNREACHABLE;
@@ -537,16 +565,16 @@ void Sieve::cross_off_count(uint64_t prime, uint64_t i)
   if (i >= wheel_.size())
     add(prime);
 
-  reset_counters();
+  reset_counter();
   Wheel& wheel = wheel_[i];
   prime /= 30;
 
   uint64_t m = wheel.multiple;
   uint64_t total_count = total_count_;
-  uint64_t counters_dist_log2 = counters_dist_log2_;
+  uint64_t counter_log2_dist = counter_.log2_dist;
   uint64_t sieve_size = sieve_.size();
-  uint32_t* counters = counters_.data();
-  uint8_t* sieve = sieve_.data();
+  uint32_t* counter = &counter_[0];
+  uint8_t* sieve = &sieve_[0];
 
   #define CHECK_FINISHED(wheel_index) \
     if_unlikely(m >= sieve_size) \
@@ -559,10 +587,12 @@ void Sieve::cross_off_count(uint64_t prime, uint64_t i)
 
   #define COUNT_UNSET_BIT(bit_index) \
     { \
-      auto is_bit = (sieve[m] >> bit_index) & 1; \
-      counters[m >> counters_dist_log2] -= is_bit; \
+      std::size_t sieve_byte = sieve[m]; \
+      std::size_t bit = sieve_byte & (1 << bit_index); \
+      std::size_t is_bit = bit >> bit_index; \
+      sieve[m] = (uint8_t) (sieve_byte - bit); \
+      counter[m >> counter_log2_dist] -= (uint32_t) is_bit; \
       total_count -= is_bit; \
-      sieve[m] &= ~(1 << bit_index); \
     }
 
   switch (wheel.index)
