@@ -13,7 +13,7 @@
 ///        method, Revista do DETUA, vol. 4, no. 6, March 2006,
 ///        pp. 759-768.
 ///
-/// Copyright (C) 2021 Kim Walisch, <kim.walisch@gmail.com>
+/// Copyright (C) 2022 Kim Walisch, <kim.walisch@gmail.com>
 ///
 /// This file is distributed under the BSD License. See the COPYING
 /// file in the top level directory.
@@ -22,20 +22,20 @@
 #include <PiTable.hpp>
 #include <primecount-internal.hpp>
 #include <fast_div.hpp>
-#include <for_atomic.hpp>
 #include <generate.hpp>
 #include <int128_t.hpp>
 #include <min.hpp>
 #include <imath.hpp>
+#include <pod_vector.hpp>
 #include <print.hpp>
+#include <RelaxedAtomic.hpp>
 #include <StatusS2.hpp>
 #include <S.hpp>
 
 #include <libdivide.h>
 #include <stdint.h>
-#include <vector>
 
-using namespace std;
+using std::numeric_limits;
 using namespace primecount;
 
 namespace {
@@ -157,22 +157,25 @@ T S2_easy_OpenMP(T x,
                  int threads,
                  bool is_print)
 {
-  // Initialize libdivide vector using primes
-  vector<libdivide::branchfree_divider<uint64_t>> lprimes(1);
-  lprimes.insert(lprimes.end(), primes.begin() + 1, primes.end());
+  // Initialize libdivide vector from primes vector
+  pod_vector<libdivide::branchfree_divider<uint64_t>> lprimes;
+  lprimes.resize(primes.size());
+  for (std::size_t i = 1; i < lprimes.size(); i++)
+    lprimes[i] = primes[i];
 
   T sum = 0;
   int64_t x13 = iroot<3>(x);
   threads = ideal_num_threads(threads, x13, 1000);
 
+  StatusS2 status(x);
   PiTable pi(y, threads);
   int64_t pi_sqrty = pi[isqrt(y)];
   int64_t pi_x13 = pi[x13];
-  int64_t min_b = max(c, pi_sqrty) + 1;
-  StatusS2 status(x);
+  RelaxedAtomic<int64_t> min_b(max(c, pi_sqrty) + 1);
 
   // for (b = pi[sqrty] + 1; b <= pi_x13; b++)
-  parallel_for_atomic_inc(min_b, b <= pi_x13)
+  #pragma omp parallel num_threads(threads) reduction(+: sum)
+  for (int64_t b = min_b++; b <= pi_x13; b = min_b++)
   {
     int64_t prime = primes[b];
     T xp = x / prime;
